/*
 * screen.c
 *
 * Written By Matthew Green, based on portions of window.c
 * by Michael Sandrof.
 *
 * Copyright(c) 1993, 1994.
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */


#include "irc.h"
#include "screen.h"
#include "status.h"
#include "window.h"
#include "output.h"
#include "vars.h"
#include "server.h"
#include "list.h"
#include "ircterm.h"
#include "names.h"
#include "ircaux.h"
#include "input.h"
#include "log.h"
#include "hook.h"
#include "dcc.h"
#include "exec.h"
#include "newio.h"
#include "misc.h"
#include "cset.h"
#ifdef TRANSLATE
#include "translat.h"
#endif
#ifdef WANT_HEBREW
#include "hebrew.h"
#endif

#ifdef WINNT
extern	int	reverse_ch;
#endif


	Window	*to_window;
	Screen	*current_screen;
	Screen	*main_screen;
	Screen	*last_input_screen;

	int	extended_handled = 0;
extern	int	in_add_to_tcl;
	
/* Our list of screens */
Screen	*screen_list = NULL;

static	void	display_lastlog_lines _((int, int, Window *));
Screen	* create_new_screen _((void));
static	void	scrollback_backwards_lines _((int));
static	void	scrollback_forwards_lines _((int));
static	char	display_highlight _((int));
static	char	display_bold _((int));
static	void	add_to_window _((Window *, unsigned char *));
static	char	*next_line_back _((Window *));

	void	add_to_tcl	_((Window *, char *));

/*
 * create_new_screen creates a new screen structure. with the help of
 * this structure we maintain ircII windows that cross screen window
 * boundaries.
 */
Screen	* create_new_screen(void)
{
	Screen	*new = NULL,
		**list;
	static	int	refnumber = 0;
	int i;
	
	for (list = &screen_list; *list; list = &((*list)->next))
	{
		if (!(*list)->alive)
		{
			new = *list;
			break;
		}
	}
	if (!new)
	{
		new = (Screen *) new_malloc(sizeof(Screen));
		memset(new, 0, sizeof(Screen));
		new->screennum = ++refnumber;
		new->next = screen_list;
		if (screen_list)
			screen_list->prev = new;
		screen_list = new;
	}
	new->last_window_refnum = 1;
	new->window_list = NULL;
	new->window_list_end = NULL;
	new->cursor_window = NULL;
	new->current_window = NULL;
	new->visible_windows = 0;
	new->window_stack = NULL;
	for (i = 0; i <= 9; i++)
		new->meta_hit[i] = 0;                   

	new->quote_hit = new->digraph_hit = new->inside_menu = 0;
	new->buffer_pos = new->buffer_min_pos = 0;
	new->input_buffer[0] = '\0';
	new->fdout = 1;
	new->fpout = stdout;
	new->fdin = 0;
#ifdef WINNT
	new->hStdin = GetStdHandle(STD_INPUT_HANDLE);
	new->hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
#endif
	if (use_input)
		FD_SET(0, &readables);
	new->fpin = stdin;
	new->alive = 1;
	new->promptlist = NULL;
	new->redirect_name = NULL;
	new->redirect_token = NULL;
	new->tty_name = NULL;
	new->li = 24;
	new->co = 79;
	new->redirect_server = -1;
	last_input_screen = new;
	return new;
}

/* 
 * add_wait_prompt:  Given a prompt string, a function to call when
 * the prompt is entered.. some other data to pass to the function,
 * and the type of prompt..  either for a line, or a key, we add 
 * this to the prompt_list for the current screen..  and set the
 * input prompt accordingly.
 */
void add_wait_prompt(char *prompt, void (*func)(char *, char *), char *data, int type)
{
	WaitPrompt **AddLoc,
		   *New;

	New = (WaitPrompt *) new_malloc(sizeof(WaitPrompt));
	New->prompt = m_strdup(prompt);
	New->data = m_strdup(data);
	New->type = type;
	New->func = func;
	New->next = NULL;
	for (AddLoc = &current_screen->promptlist; *AddLoc;
			AddLoc = &(*AddLoc)->next);
	*AddLoc = New;
	if (AddLoc == &current_screen->promptlist)
		change_input_prompt(1);
}

void set_current_screen(Screen *screen)
{
	current_screen = (screen && screen->alive) ? screen : screen_list; 
}

/*
 * window_redirect: Setting who to non null will cause IRCII to echo all
 * commands and output from that server (including output generated by IRCII)
 * to who.  Setting who to null disables this 
 */
void window_redirect(char *who, int server)
{
	char	buf[128];

	if (who)
		snprintf(buf, 120, "%04d%s", server, who);
	else
		snprintf(buf, 120, "%04d#R#", server);
	malloc_strcpy(&current_screen->redirect_token, buf);
	malloc_strcpy(&current_screen->redirect_name, who);
	current_screen->redirect_server = server;
}

int check_screen_redirect(char *nick)
{
	Screen	*screen,
		*tmp_screen;
	
	for (screen = screen_list; screen; screen = screen->next)
	{
		if (!screen->redirect_token)
			continue;
		if (!strncmp(nick, screen->redirect_token, strlen(screen->redirect_token)))
		{
			tmp_screen = current_screen;
			current_screen = screen;
			window_redirect(NULL, from_server);
			current_screen = tmp_screen;
			return 1;
		}
	}
	return 0;
}

#ifdef WINDOW_CREATE
void kill_screen(Screen *screen)
{
	Window	*window;

	if (main_screen == screen)
	{
		say("You may not kill the main screen");
		return;
	}
	if (screen->fdin)
	{
		if (use_input)
			FD_CLR(screen->fdin, &readables);
		close(screen->fdout);
		close(screen->fdin);
	}
	while ((window = screen->window_list))
	{
		screen->window_list = window->next;
		add_to_invisible_list(window);
	}

	if (last_input_screen == screen)
		last_input_screen = screen_list;

	say("The screen is now dead.");
	screen->alive = 0;
}

int is_main_screen(Screen *screen)
{
	return (screen == main_screen);
}

#else

static	int is_main_screen(Screen *screen)
{
	return 1;
}

#endif /* WINDOW_CREATE */

/*
 * scroll_window: Given a pointer to a window, this determines if that window
 * should be scrolled, or the cursor moved to the top of the screen again, or
 * if it should be left alone. 
 */
void scroll_window(Window *window)
{
	if (dumb)
		return;
	if (window->cursor == window->display_size)
	{
		if (window->scroll)
		{
			int	scroll,
			i;

			if ((scroll = get_int_var(SCROLL_LINES_VAR)) <= 0)
				scroll = 1;

			for (i = 0; i < scroll; i++)
			{
				new_free (&window->top_of_display->line);
				window->top_of_display =
					window->top_of_display->next;
			}
			if (window->visible)
			{
				if (window->status_split)
				{
					term_scroll(window->top+window->menu.lines,
						window->top + window->menu.lines +
						window->cursor - 1, scroll);
					window->cursor -= scroll;
					term_move_cursor(0, window->cursor + window->top + window->menu.lines);
				}
				else
				{
					term_scroll(window->top,
						window->top - window->menu.lines +
						window->cursor, scroll);
					window->cursor -= scroll;
					term_move_cursor(0, window->cursor + window->top);
				}
			}
			else
				window->cursor -= scroll;
		}
		else
		{
			window->cursor = 0;
			if (window->visible)
				if (window->status_split)
					term_move_cursor(0, window->top + window->menu.lines/*+window->status_split*/);
				else
					term_move_cursor(0, window->top);
		}
	}
	else if (window->visible && current_screen->cursor_window == window)
	{
#ifdef WINNT
		NT_MoveToLineOrChar(current_screen->hStdout, 1, 1);
		NT_MoveToLineOrCharAbs(current_screen->hStdout, 0, 0);
#else
		term_cr();
		term_newline();
#endif
	}
	if (window->visible && current_screen->cursor_window)
		term_clear_to_eol();
}

#ifdef WINNT
/* display_highlight: turns off and on the display highlight.  */
static	char display_highlight(int flag)
{
	if (flag == reverse_ch)
		return (flag);
	switch (flag)
	{
		case ON:
		{
			if (get_int_var(INVERSE_VIDEO_VAR))
                        	reverse_ch = 1;
			flag = OFF;
                	break;
		}
		case OFF:
		{
			if (get_int_var(INVERSE_VIDEO_VAR))
                        	reverse_ch = 0;
			flag = ON;
                	break;
		}
		case TOGGLE:
		{
			if (reverse_ch)
			{
				if (get_int_var(INVERSE_VIDEO_VAR))
					reverse_ch = 0;
				flag = ON;
			}
			else
			{
				if (get_int_var(INVERSE_VIDEO_VAR))
					reverse_ch = 1;
				flag = OFF;
			}
		}
	}
	set_term_inverse(reverse_ch);
	return flag;
}

/* display_bold: turns off and on the display bolding.  */
static	char display_bold(int flag)
{
	if (flag == intensity)
		return (flag);
	switch (flag)
	{
	case ON:
		if (get_int_var(BOLD_VIDEO_VAR))
			intensity = 1;
		flag = OFF;
        	break;
	case OFF:
		if (get_int_var(BOLD_VIDEO_VAR))
			intensity = 0;
               	break;
		flag = ON;
	case TOGGLE:
		if (intensity)
		{
			if (get_int_var(BOLD_VIDEO_VAR))
				intensity = 0;
			flag = ON;
		}
		else
		{
			if (get_int_var(BOLD_VIDEO_VAR))
				intensity = 1;
			flag = OFF;
		}
	}
	set_term_intensity(intensity);
	return flag;
}

/* display_bold: turns off and on the display bolding.  */
static char display_underline (int flag)
{
	if (flag == underline)
		return (flag);

	switch (flag)
	{
		case ON:
		{
			if (get_int_var(UNDERLINE_VIDEO_VAR))
				underline = 1;
			flag = OFF;
	               	break;
		}
		case OFF:
		{
			if (get_int_var(UNDERLINE_VIDEO_VAR))
				underline = 0;
			flag = ON;
	               	break;
		}
		case TOGGLE:
		{
			if (underline)
			{
				if (get_int_var(UNDERLINE_VIDEO_VAR))
					underline = 0;
				flag = ON;
			}
			else
			{
				if (get_int_var(UNDERLINE_VIDEO_VAR))
					underline = 1;
				flag = OFF;
			}
		}
	}
	set_term_underline(underline);
	return flag;
}

#else
/* display_highlight: turns off and on the display highlight.  */
static	char display_highlight(int flag)
{
	static	int	highlight = OFF;

	if (flag == highlight)
		return (flag);
	switch (flag)
	{
		case ON:
		{
			highlight = ON;
			if (get_int_var(INVERSE_VIDEO_VAR))
				term_standout_on();
			return (OFF);
		}
		case OFF:
		{
			highlight = OFF;
			if (get_int_var(INVERSE_VIDEO_VAR))
				term_standout_off();
			return (ON);
		}
		case TOGGLE:
		{
			if (highlight == ON)
			{
				highlight = OFF;
				if (get_int_var(INVERSE_VIDEO_VAR))
					term_standout_off();
				return (ON);
			}
			else
			{
				highlight = ON;
				if (get_int_var(INVERSE_VIDEO_VAR))
					term_standout_on();
				return (OFF);
			}
		}
	}
	return flag;
}
 
/* display_bold: turns off and on the display bolding.  */
static	char display_bold(int flag)
{
	static	int	bold = OFF;

	if (flag == bold)
		return (flag);
	switch (flag)
	{
	case ON:
		bold = ON;
		if (get_int_var(BOLD_VIDEO_VAR))
			term_bold_on();
		return (OFF);
	case OFF:
		bold = OFF;
		if (get_int_var(BOLD_VIDEO_VAR))
			term_bold_off();
		return (ON);
	case TOGGLE:
		if (bold == ON)
		{
			bold = OFF;
			if (get_int_var(BOLD_VIDEO_VAR))
				term_bold_off();
			return (ON);
		}
		else
		{
			bold = ON;
			if (get_int_var(BOLD_VIDEO_VAR))
				term_bold_on();
			return (OFF);
		}
	}
	return OFF;
}

/* display_bold: turns off and on the display bolding.  */
static char display_underline (int flag)
{
	static	int	underline = OFF;

	if (flag == underline)
		return (flag);

	switch (flag)
	{
		case ON:
		{
			underline = ON;
			if (get_int_var(UNDERLINE_VIDEO_VAR))
				term_underline_on();
			return (OFF);
		}
		case OFF:
		{
			underline = OFF;
			if (get_int_var(UNDERLINE_VIDEO_VAR))
				term_underline_off();
			return (ON);
		}
		case TOGGLE:
		{
			if (underline == ON)
			{
				underline = OFF;
				if (get_int_var(UNDERLINE_VIDEO_VAR))
					term_underline_off();
				return (ON);
			}
			else
			{
				underline = ON;
				if (get_int_var(UNDERLINE_VIDEO_VAR))
					term_underline_on();
				return (OFF);
			}
		}
	}
	return OFF;
}
#endif

/*
 * This actually outputs a string to the screen.  Youll notice this function
 * is a lot simpler than it was before, because all of that busiwork it did
 * before was completely bogus.  What happens now is:
 *
 * Any time we get a recognized highlight character, we handle it by 
 * toggling the appropriate setting.  This might waste a few function calls
 * if people have two of the same highlight char next to each other, but
 * that savings isnt worth the maintainance nightmare of the previous code.
 *
 * Any other character that isnt a recognized highlight character is 
 * dispatched to the screen.  Since we use term_putchar(), any special
 * handling of ansi sequences is filtered there.  We do not know, and we
 * do not care about what is actually in the string.  We can (and do) assume
 * that the string passed to us in "str" will fit on the screen at whatever
 * position the cursor is at.
 *
 * Unless the last character on the line is an ALL_OFF character, output_line
 * will *NOT* reset any attributes that may be set during the source of the
 * output.  This is so you can output one or more lines that constitute part
 * of a "logical" line, and any highlights or colors will "bleed" onto the
 * next physical line.  The last such physical line in a logical line without
 * exception MUST end with the ALL_OFF character, or everything will go all
 * higgledy-piggledy.  As long as you do that, everything will look ok.
 */
int output_line (register unsigned char *str)
{
register unsigned char	*ptr = str;
	int	beep = 0;
#ifdef WINNT
	int	len = strlen(str);
#endif
	/* Walk the string looking for special chars. */
#ifdef WINNT
	nt_write(current_screen->hStdout,str, len);
#else

	while (*ptr)
	{
		switch (*ptr)
		{
			case REV_TOG:
			{
				display_highlight(TOGGLE);
				break;
			}
			case UND_TOG:
			{
				display_underline(TOGGLE);
				break;
			}
			case BOLD_TOG:
			{
				display_bold(TOGGLE);
				break;
			}
			case ALL_OFF:
			{
				display_underline(OFF);
				display_highlight(OFF);
				display_bold(OFF); /* This turns it all off */

				if (!ptr[1])	/* We're the end of line */
					term_bold_off();
				break;
			}
			case '\007':
			{
				beep++;
				break;
			}
			default:
			{
#ifdef WINNT
				nt_write(current_screen->hStdout, ptr, 1);
#else
				term_putchar(*ptr);
#endif
				break;
			}
		}
		ptr++;
	}


	if (beep)
		term_beep();
#endif
	term_clear_to_eol();
	return 0;
}



/*
 * cursor_not_in_display: This forces the cursor out of the display by
 * setting the cursor window to null.  This doesn't actually change the
 * physical position of the cursor, but it will force rite() to reset the
 * cursor upon its next call 
 */
void cursor_not_in_display _((void))
{
	current_screen->cursor_window = NULL;
}

/*
 * cursor_in_display: this forces the cursor_window to be the
 * current_screen->current_window. 
 * It is actually only used in hold.c to trick the system into thinking the
 * cursor is in a window, thus letting the input updating routines move the
 * cursor down to the input line.  Dumb dumb dumb 
 */
void cursor_in_display _((void))
{
	current_screen->cursor_window = curr_scr_win;
}

/*
 * is_cursor_in_display: returns true if the cursor is in one of the windows
 * (cursor_window is not null), false otherwise 
 */
int is_cursor_in_display(Screen *screen)
{
	return ((screen ? screen : current_screen)->cursor_window ? 1 : 0);
}

void redraw_resized(Window *window, ShrinkInfo Info, int AnchorTop)
{
	if (!AnchorTop)
	{
		if (Info.bottom < 0)
			if (window->status_split)
			term_scroll(window->top+window->menu.lines/*+window->status_split*/+Info.bottom,
				window->top + window->menu.lines/* + window->status_split*/ +
				window->display_size - 1, Info.bottom);
			else
			term_scroll(window->top+Info.bottom,
				window->top + window->display_size - 1, Info.bottom);
		else if (Info.bottom)
			if (window->status_split)
			term_scroll(window->top+window->menu.lines/* + window->status_split*/,
				window->top + window->menu.lines/* + window->status_split*/ +
				window->display_size -1, Info.bottom);
			else
			term_scroll(window->top,
				window->top + window->display_size -1, Info.bottom);
	}
}

/*
 * resize_display: After determining that the screen has changed sizes, this
 * resizes all the internal stuff.  If the screen grew, this will add extra
 * empty display entries to the end of the display list.  If the screen
 * shrank, this will remove entries from the end of the display list.  By
 * doing this, we try to maintain as much of the display as possible. 
 *
 * This has now been improved so that it returns enough information for
 * redraw_resized to redisplay the contents of the window without having
 * to redraw too much.
 */
ShrinkInfo resize_display(Window *window)
{
	int	cnt,
		i;
	Display *tmp, *pre_ip;
	int	Wrapped = 0;
	ShrinkInfo Result;

	Result.top = Result.bottom = 0;
	Result.position = window->cursor;
	if (dumb)
		return Result;

	if (!window->top_of_display)
	{
		window->top_of_display = (Display *)new_malloc(sizeof(Display));
		window->top_of_display->line = NULL;
		window->top_of_display->linetype = LT_UNLOGGED;
		window->top_of_display->next = window->top_of_display;
		window->display_ip = window->top_of_display;
		window->old_size = 1;
	}
	/* cnt = size - window->display_size; */
	cnt = window->display_size - window->old_size;
	if (cnt > 0)
	{
		Display *new = NULL;

	/*
	 * screen got bigger: got to last display entry and link in new
	 * entries 
	 */
		for (tmp = window->top_of_display, i = 0; i < (window->old_size - 1); i++, tmp = tmp->next)
			;
		for (i = 0; i < cnt; i++)
		{
			new = (Display *) new_malloc(sizeof(Display));
			new->line = NULL;
			new->linetype = LT_UNLOGGED;
			new->next = tmp->next;
			tmp->next = new;
		}
		if (window->display_ip == window->top_of_display &&
		    window->top_of_display->line)
			window->display_ip = new;
		Result.top = 0;
		Result.bottom = cnt;
		Result.position = 0;
	}
	else if (cnt < 0)
	{
		Display *ptr;

	/*
	 * screen shrank: find last display entry we want to keep, and remove
	 * all after that point 
	 */
		cnt = -cnt;
		for (pre_ip = window->top_of_display;
		    pre_ip->next != window->display_ip;
		    pre_ip = pre_ip->next)
		        ;

		for (tmp = pre_ip->next, i =0; i < cnt; i++, tmp = ptr)
		{
			ptr = tmp->next;
			if (tmp == window->top_of_display)
			{
				if (tmp->line)
					Wrapped = 1;
				window->top_of_display = ptr;
			}
			if (Wrapped)
				Result.top--;
			else
				Result.bottom--;
			new_free(&(tmp->line));
			new_free((char **)&tmp);
		}
		window->display_ip = pre_ip->next = tmp;
		window->cursor += Result.top;
		if (!window->scroll)
		{
			if (window->cursor == window->display_size)
				window->cursor = 0;
			new_free(&window->display_ip->line);
		}
	}
	window->update |= REDRAW_DISPLAY_FULL | REDRAW_STATUS;
	window->old_size = window->display_size;
	return Result;
}

/*
 * recalculate_windows: this is called when the terminal size changes (as
 * when an xterm window size is changed).  It recalculates the sized and
 * positions of all the windows.  Currently, all windows are rebalanced and
 * window size proportionality is lost 
 */
void recalculate_windows _((void))
{
	int	base_size,
	size,
	top,
	extra;
	Window	*tmp = current_screen->window_list;
	
	if (dumb)
		return;

	base_size = ((LI - 1) / current_screen->visible_windows) - 1;
	extra = (LI - 1) - ((base_size + 1) * current_screen->visible_windows);
	top = 0;

	for (tmp = current_screen->window_list; tmp; tmp = tmp->next)
	{
		tmp->update |= REDRAW_DISPLAY_FULL | REDRAW_STATUS;
		if (extra)
		{
			extra--;
			size = base_size + 1;
		}
		else
			size = base_size;

		tmp->display_size = size - tmp->menu.lines - tmp->double_status;

		if (tmp->display_size <= 0)
			tmp->display_size = 1;
		tmp->top = top;
		tmp->bottom = top + size - tmp->double_status;
		if (!tmp->status_split)
			tmp->bottom = tmp->bottom - 1;
		top += size + 1;
	}
}

/*
 * clear_window: This clears the display list for the given window, or
 * current window if null is given.  
 */
void clear_window(Window *window)
{
	int	i,
		cnt;

	if (dumb)
		return;
	if (window == NULL)
		window = curr_scr_win;
	erase_display(window);
	if (window->status_split)
	{
		term_move_cursor(0, window->top + window->menu.lines/* + window->status_split*/);
		cnt = window->bottom - window->top - window->menu.lines;
	} 
	else
	{
		term_move_cursor(0, window->top);
		cnt = window->bottom - window->top;
	}
	for (i = 0; i < cnt; i++)
	{
		term_clear_to_eol();
#ifdef WINNT
		nt_write(current_screen->hStdout, "\n", 1);
#else
		term_newline();
#endif
	}
	term_flush();
}

/* clear_all_windows: This clears all *visible* windows */
void clear_all_windows(int unhold)
{
	Window	*tmp;

	for (tmp = current_screen->window_list; tmp; tmp = tmp->next)
	{
		if (unhold)
			hold_mode(tmp, OFF, 1);
		clear_window(tmp);
	}
}

/*
 * redraw_window: This redraws the display list for the given window. Some
 * special considerations are made if you are just redrawing one window as
 * opposed to using this routine to redraw the whole screen one window at a
 * time 
 *
 * A negative value in just_one indicates not all of the window needs to
 * be redrawn.
 */
void redraw_window(Window *window, int just_one)
{
	Display *tmp;
	int	i;
	int	StartPoint;
	int	yScr;

	if (dumb || !window->visible)
		return;
	window = window ? window : curr_scr_win;

	if (just_one < 0)
	{
	/* This part of the window is scrolling into view */
		StartPoint = -just_one;
		just_one = 0;
	}
	else
	{
		StartPoint = 0;
		if (window->scrolled_lines)
			display_lastlog_lines(window->scrolled_lines-window->display_size + 1, window->scrolled_lines, window);
	}
	if (window->scrolled_lines + StartPoint < window->display_size)
		yScr = window->scrolled_lines + StartPoint;
	else
		yScr = 0;
	if (window->status_split)
		term_move_cursor(0, window->top+window->menu.lines+yScr);
	else
		term_move_cursor(0, window->top+yScr);
	for (tmp = window->top_of_display, i = 0; i < window->display_size-window->scrolled_lines; i++, tmp = tmp->next)
	{
		if (i < StartPoint)
			continue;
		if (tmp->line)
			rite(window, tmp->line, 1, 1, 0, 0);
		else
		{
			if (just_one)
				term_clear_to_eol();
#ifdef WINNT
			nt_write(current_screen->hStdout, "\n", 1);
#else
			term_newline();
#endif
		}
	}
	term_flush();
}

/*
 * recalculate_window_positions: This runs through the window list and
 * re-adjusts the top and bottom fields of the windows according to their
 * current positions in the window list.  This doesn't change any sizes of
 * the windows 
 */
void recalculate_window_positions(void)
{
	Window	*tmp = current_screen->window_list;
	int	top;

	top = 0;
	for (tmp = current_screen->window_list; tmp; tmp = tmp->next)
	{
		tmp->update |= REDRAW_DISPLAY_FULL | REDRAW_STATUS;
		tmp->top = top;
		if (tmp->status_split)
		{
			tmp->bottom = top + tmp->display_size + tmp->menu.lines;
			top += tmp->display_size + tmp->menu.lines + 1 + tmp->double_status;
		} 
		else
		{
			tmp->bottom = top + tmp->display_size + tmp->menu.lines;
			top += tmp->display_size + tmp->menu.lines + 1 + tmp->double_status;
		}
	}
}

/*
 * redraw_all_windows: This basically clears and redraws the entire display
 * portion of the screen.  All windows and status lines are draws.  This does
 * nothing for the input line of the screen.  Only visible windows are drawn 
 */
void redraw_all_windows(void)
{
	Window	*tmp;

	if (dumb)
		return;
	for (tmp = current_screen->window_list; tmp; tmp = tmp->next)
		tmp->update = REDRAW_STATUS | REDRAW_DISPLAY_FAST;
}

/*
 * split_up_line -- converts a logical line into one or more physical lines
 * accounting for special characters, unprintables, etc.  This function
 * is large and obtuse, so itll take a bit of explaining.
 *
 * Essentialy two copies of the string are made: The original string (str)
 * is walked and is copied byte for byte into "buffer", with extra characters
 * being inserted for the sake of special control characters (like beeps or
 * tabs or whatever).  As each line in "buffer" is created, it is malloc-
 * copied off into "output" which holds the final broken result.  When the
 * entire input has been converted into "buffer" and then copied off into
 * "output", the "output" is returned, and passed off to output_line.
 *
 * As much as possible, i have had tried to avoid extraneous malloc() and
 * free calls which are essentially a waste of CPU time.  Note that this
 * means that an exceptionally long line will not have all of its lines
 * free()d until another line of such length is processed.  That can mean
 * a small amount of space that is not "lost" but will not be reclaimed, 
 * which should not be terribly significant, since it saves a tremendous
 * amount of CPU.
 *
 * Note that there are no limiations on the put any longer.  Neither the
 * size of "buffer" (2048 characters) nor the size of "output" (40 lines)
 * are completely inflexible any longer.  The only true limitation now is
 * a screen width of 2048 characters (absurdly large).
 *
 * XXXX -- If it werent for scrollback, you could roll output_line into this
 * function and simplify the output chain greatly, without having to place
 * arbitrary limits on the output.
 */
#define	MAXIMUM_SPLITS	40
#define MAX_SCREEN_BUFFER (BIG_BUFFER_SIZE * 2)
unsigned char **split_up_line (unsigned char *str)
{
static	unsigned char	**output = NULL;
static	int		output_size = 0;
	unsigned char	buffer[BIG_BUFFER_SIZE + 1];
const	unsigned char	*ptr;
	unsigned char	*cont_ptr,
			*cont = empty_string,
			c;
	int		pos = 0,		/* Current pos in "buffer" */
			col = 0,		/* Current col on display */
			word_break = 0,		/* Last end of word */
			i,			/* Generic counter */
			indent = 0,		/* Start of 2nd word */
			beep_cnt = 0,		/* Controls number of beeps */
			beep_max,		/* what od you think? */
			tab_cnt = 0,
			tab_max,
			line = 0;		/* Current pos in "output" */
	int		len;			/* Used for counting tabs */
	unsigned char	*pos_copy;		/* Used for reconstruction */
	int		do_indent;
	int		newline = 0;

	/* XXXXX BOGUS! XXXXX */
	if (!output_size)
	{
		int new = MAXIMUM_SPLITS;
		RESIZE(output, char *, new);
		output_size = new;
	}

	*buffer = 0;

	/* Especially handle blank lines */
	if (!*str)
		str = " ";


	beep_max = get_int_var(BEEP_VAR) ? get_int_var(BEEP_MAX_VAR) > 5 ? 5 :get_int_var(BEEP_MAX_VAR) : -1;
	tab_max = get_int_var(TAB_VAR) ? get_int_var(TAB_MAX_VAR) : -1;
	do_indent = get_int_var(INDENT_VAR);

	if (!(cont_ptr = get_string_var(CONTINUED_LINE_VAR)))
		cont_ptr = empty_string;


	/*
	 * Walk the entire input string and convert it as neccesary.
	 * Intermediate results go into 'buffer', and the final destination
	 * is 'output'.
	 */
#ifdef WANT_HEBREW
	/*
	 * crisk: Hebrew processing the line.
	 * 
	 */
	if (get_int_var(HEBREW_TOGGLE_VAR))
		hebrew_process(str);
#endif

	for (ptr = str; *ptr && (pos < 3 * BIG_BUFFER_SIZE - 20); ptr++)
	{
#ifdef TRANSLATE
		if (translation)
			*ptr = transToClient[*ptr];
#endif
		switch (*ptr)
		{
			case '\007':	/* bell */
			{
				beep_cnt++;
				if (beep_max == -1 || beep_cnt > beep_max)
				{
					buffer[pos++] = REV_TOG;
					buffer[pos++] = ((*ptr) & 127) | 64;
					buffer[pos++] = REV_TOG;
					col++;
				}
				else 
					buffer[pos++] = *ptr;
				break;
			}
			case '\011':	/* tab */
			{
				tab_cnt++;
				if (tab_max > 0 && tab_cnt > tab_max)
				{
					buffer[pos++] = REV_TOG;
					buffer[pos++] = ((*ptr) & 127) | 64;
					buffer[pos++] = REV_TOG;
					col++;
				}
				else
				{
					if (indent == 0)
						indent = -1;
					word_break = pos;

					/*
					 * Tabs only go as far as the right
					 * edge of the screen -- no farther
					 */
					len = 8 - (col % 8);
					for (i = 0; i < len; i++)
					{
						buffer[pos++] = ' ';
						if (col++ >= CO)
							break;
					}
				}
				break;
			}
			case '\n':	/* Forced newline */
			{
				newline = 1;
				if (indent == 0)
					indent = -1;
				word_break = pos;
				break;
			}
			case ' ':	/* word break */
			{
				if (indent == 0)
					indent = -1;
				word_break = pos;
				buffer[pos++] = *ptr;
				col++;
				break;
			}
			case UND_TOG:
			case ALL_OFF:
			case REV_TOG:
			case BOLD_TOG:
			{
				buffer[pos++] = *ptr;
				break;
			}
			case '\003':	/* ^C colors */
			{
				buffer[pos++] = *ptr;
				if (ptr[1] && isdigit(ptr[1]))
				{
					while (ptr[1] && isdigit(ptr[1]))
					{
						buffer[pos++] = ptr[1];
						ptr++, i++;
					}

					if (ptr[1] == ',')
					{
						buffer[pos++] = ptr[1];
						ptr++;
						while (ptr[1] && isdigit(ptr[1]))
						{
							buffer[pos++] = ptr[1];
							ptr++, i++;
						}
					}
				}
				break;
			}
			case '\033':	/* ESCape */
			{
				/* 
				 * This is the only character that
				 * term_putchar lets go by unchecked.
				 * so we check it here.  Everything
				 * else doesnt need to be filtered.
				 */
				if (!get_int_var(DISPLAY_ANSI_VAR))
				{
					buffer[pos++] = REV_TOG;
					buffer[pos++] = '[';
					buffer[pos++] = REV_TOG;
					col++;
					break;
				}
				/* ELSE FALLTHROUGH */
			}

			/* Everything else including unfiltered ESCapes */
			default:
			{
				/* Check if printable... */
				if (!vt100_decode(*ptr))
				{
					if (indent == -1)
						indent = col;
					col++;
				}

				buffer[pos++] = *ptr;
				break;
			}
		}


#if not_needed
		if (pos == BIG_BUFFER_SIZE)
			*ptr = '\0';
#endif

		/*
		 * Ok.  This test is to see if we've reached the right side
		 * of the display.  If we do, then we have three tasks:
		 *  1. Break off the current line at the end of the last word
		 *  2. Start the new line with the "continued line" string
		 *  3. Put back the stuff already parsed after the break.
		 */
		if (col >= CO || newline)
		{
			/* one big long line, no word breaks */
			if (word_break == 0)
				word_break = pos;

			/*
			 * Resize output if neccesary.
			 * XXX Note that this uses -3, so that there are
			 * always at least two lines available when we
			 * exit the for loop -- one for the last line,
			 * and one for the 
			 */
			if (line >= output_size - 3)
			{
				int new = output_size + MAXIMUM_SPLITS + 1;
				RESIZE(output, char *, new);
				output_size = new;
			}

			/* 1. Break off the current line at the end */
			/*
			 * Copy the current line into the return values.
			 */
			c = buffer[word_break];
			buffer[word_break] = '\0';
			malloc_strcpy((char **)&(output[line++]), buffer);
			buffer[word_break] = c;

		


			/* 2. Calculate the "continued line" string */

			/*
			 * If the 'continued line' string has not yet
			 * been set, we need to set it here (since at
			 * this point the first line is done, and we cant
			 * do the /SET INDENT until we know how the first
			 * line is layed out.
			 */
			if (!*cont && do_indent && (indent < CO / 3) && 
				(strlen(cont_ptr) < indent))
			{
				cont = alloca(indent + 2);
				sprintf(cont, "%-*s", indent, cont_ptr);
			}
			/*
			 * If cont_ptr has not been initialize, do so.
			 * Otherwise we dont do anything.
			 */
			else if (!*cont && *cont_ptr)
				cont = cont_ptr;


			/* 3. Paste the new line back together again. */

			/*
			 * Skip over the spaces that comprised the break
			 * between the lines.
			 */
			while (buffer[word_break] == ' ' && word_break <= pos)
				word_break++;

			/*
			 * Now re-create the new line
			 * and figure out where we are.
			 */
			buffer[pos] = 0;
			pos_copy = alloca(strlen(buffer + word_break) + 2);
			strcpy(pos_copy, buffer + word_break);

			strcpy(buffer, cont);
			strcat(buffer, pos_copy);
			col = pos = strlen(buffer);
			word_break = 0;
			newline = 0;

			if (get_int_var(DISPLAY_ANSI_VAR))
				col -= count_ansi(buffer, -1);
		}
	}

	/*
	 * At this point we've completely walked the input source, so we 
	 * clean up anything left in the intermediate buffer.  This is the 
	 * "typical" case, since most output is only one line long.
	 */
	buffer[pos++] = ALL_OFF;
	buffer[pos] = '\0';
	if (*buffer)
		malloc_strcpy((char **)&(output[line++]), buffer);

	new_free(&output[line]);
	return output;
}
 
/*
 * rite: this routine displays a line to the screen adding bold facing when
 * specified by ^Bs, etc.  It also does handle scrolling and paging, if
 * SCROLL is on, and HOLD_MODE is on, etc.  This routine assumes that str
 * already fits on one screen line.  If show is true, str is displayed
 * regardless of the hold mode state.  If redraw is true, it is assumed we a
 * redrawing the screen from the display_ip list, and we should not add what
 * we are displaying back to the display_ip list again. 
 *
 * Note that rite sets display_highlight() to what it was at then end of the
 * last rite().  Also, before returning, it sets display_highlight() to OFF.
 * This way, between susequent rites(), you can be assured that the state of
 * bold face will remain the same and the it won't interfere with anything
 * else (i.e. status line, input line). 
 */
int rite(Window *window, unsigned char *str, int show, int redraw, int backscroll, int logged)
{
static	int	high = OFF;
	Screen	*old_current_screen = current_screen;

	if (!backscroll && window->scrolled_lines)
		window->new_scrolled_lines++;

	if (window->hold_mode && window->hold_on_next_rite && !redraw && !backscroll)
	{
		/* stops output from going to the window */
		window->hold_on_next_rite = 0;
		hold_mode(window, ON, 1);
		if (show)
			return (1);
	}
	/*
	 * Don't need to worry about the current_screen if the window isn't
	 * visible, as hidden windows aren't attached to a screen anyway
	 */
	if (window->visible)
	{
		old_current_screen = current_screen;
		set_current_screen(window->screen);
	}

	if (!show && (hold_output(window) || hold_queue(window)) && !redraw && !backscroll)
		/* sends window output to the hold list for that window */
		add_to_hold_list(window, str, logged);
	else
	{
		if (!redraw && !backscroll)
		{
		/*
		 * This isn't a screen refresh, so add the line to the display
		 * list for the window 
		 */
			if (window->scroll)
				scroll_window(window);

			malloc_strcpy(&(window->display_ip->line), str);
			window->display_ip->linetype = logged;
			window->display_ip = window->display_ip->next;

			if (!window->scroll)
				new_free(&window->display_ip->line);
		}

		/*
		 * XXXX - this probably doesnt belong here, but so what.
		 * archon wanted a way to have hidden windows beep, even
		 * though theyre hidden.  This is a relatively minor hack,
		 * hopefully it wont offend too many people. >;-)
		 */
		if (!window->visible && window->beep_always && index(str, '\007'))
		{
			Window *old_to_window;

			term_beep();
			old_to_window = to_window;
			to_window = curr_scr_win;
			say("Beep in window %d", window->refnum);
			to_window = old_to_window;
		}

		if (window->visible)
		{
			/* make sure the cursor is in the appropriate window */
			if (current_screen->cursor_window != window && !redraw && !backscroll)
			{
				current_screen->cursor_window = window;
				if (!window->status_split && window->menu.lines)
					term_move_cursor(0, window->cursor + window->top);
				else
					term_move_cursor(0, window->cursor + window->top + window->menu.lines);
			}
			display_highlight(high);
			output_line(str);
		}
		else if (!(window->miscflags & WINDOW_NOTIFIED))
		{
			if (who_level & window->notify_level)
			{
				window->miscflags |= WINDOW_NOTIFIED;
				if (window->miscflags & WINDOW_NOTIFY)
				{
					Window	*old_to_window;
					int	lastlog_level;

					lastlog_level =	set_lastlog_msg_level(LOG_CRAP);
					old_to_window = to_window;
					to_window = curr_scr_win;
					say("Activity in window %d", window->refnum);
					to_window = old_to_window;
					set_lastlog_msg_level(lastlog_level);
				}
/*				update_all_status(curr_scr_win, NULL, 0);*/
			}
		}

		if (!redraw && !backscroll)
		{
			window->cursor++;
			window->line_cnt++;
			if (window->scroll)
			{
				if (window->line_cnt >= window->display_size)
				{
					window->hold_on_next_rite = 1;
					window->line_cnt = 0;
				}
			}
			else
			{
				scroll_window(window);
				if (window->cursor == (window->display_size -1))
					window->hold_on_next_rite = 1;
			}

			if (window->visible)
				high = display_highlight(OFF);
		}
		else if (window->visible)
		{
			high = display_highlight(OFF);
#ifdef WINNT
			NT_MoveToLineOrChar(current_screen->hStdout, 1, 1);
			NT_MoveToLineOrCharAbs(current_screen->hStdout, 0, 0);
#else
			term_cr();
			term_newline();
#endif
		}
	}
	if (window->visible)
		set_current_screen(old_current_screen);
	return (0);
}


/*
 * add_to_window: adds the given string to the display.  No kidding. This
 * routine handles the whole ball of wax.  It keeps track of what's on the
 * screen, does the scrolling, everything... well, not quite everything...
 * The CONTINUED_LINE idea thanks to Jan L. Peterson (jlp@hamblin.byu.edu)  
 *
 * At least it used to. Now most of this is done by split_up_line, and this
 * function just dumps it onto the screen. This is because the scrollback
 * functions need to be able to figure out how to split things up too.
 */
static	void add_to_window(Window *window, unsigned char *str)
{

	if (do_hook(WINDOW_LIST, "%u %s", window->refnum, str))
	{
		unsigned char	**lines;
		int	logged;

		add_to_log(window->log_fp, 0, str);
		add_to_lastlog(window, str);
		display_highlight(OFF);
		display_bold(OFF);
		logged = islogged(window);

		for (lines = split_up_line(str); *lines; lines++)
		{
			rite(window, *lines, 0, 0, 0, logged);
			if (logged == 1)
				logged = 2;
		}
		update_all_status(curr_scr_win, NULL, 0);
		term_flush();
	}
}

/*
 * add_to_screen: This adds the given null terminated buffer to the screen.
 * That is, it determines which window the information should go to, which
 * lastlog the information should be added to, which log the information
 * should be sent to, etc 
 */
void add_to_screen(unsigned char *buffer)
{
	Window	*tmp = NULL;
	int	flag;


	if (!get_int_var(DISPLAY_ANSI_VAR))
		strcpy(buffer, stripansicodes(buffer));

	if (!current_screen || !curr_scr_win)
	{
		puts(buffer);
		return;
	}
#ifdef WANT_TCL
	if (in_add_to_tcl)
	{
		add_to_tcl(tmp, buffer);
		return;
	}
#endif
	/* Handles output redirection first */
	if (current_screen->redirect_name && from_server == current_screen->redirect_server)
		send_text(current_screen->redirect_name, buffer, NULL, 0, 0);

	if (dumb)
	{
		add_to_lastlog(curr_scr_win, buffer);
		if (do_hook(WINDOW_LIST, "%u %s", curr_scr_win->refnum, buffer))
			puts(buffer);
		fflush(current_screen->fpout);
		return;
	}
	if (in_window_command)
		update_all_windows();

	if ((who_level == LOG_CURRENT) && (curr_scr_win->server == from_server))
	{
		add_to_window(curr_scr_win, buffer);
		return;
	}
	else if (to_window)
	{
		add_to_window(to_window, buffer);
		return;
	}
	else if (who_from && *who_from)
	{
		flag = 1;
		while ((tmp = traverse_all_windows(&flag)) != NULL)
		{
			if (((from_server == tmp->server) || (from_server == -1)) &&
			    (who_level & tmp->window_level))
			{
				add_to_window(tmp, buffer);
				return;
			}
			if (tmp->current_channel)
			{
				if (!my_stricmp(who_from, tmp->current_channel) && 
					tmp->server == from_server)
				{
					add_to_window(tmp, buffer);
					return;
				}
			} 
			if (tmp->query_nick &&
				(((who_level == LOG_MSG || who_level == LOG_NOTICE)
				&& !my_stricmp(who_from, tmp->query_nick) &&
				from_server == tmp->server) ||
				(who_level == LOG_DCC &&
				(*tmp->query_nick == '=' || *tmp->query_nick == '-') &&
				!my_stricmp(who_from, tmp->query_nick + 1))))
			{
				add_to_window(tmp, buffer);
				return;
			}
			if (from_server == tmp->server)
			{
				if (find_in_list((List **)&(tmp->nicks), who_from, !USE_WILDCARDS))
				{
					add_to_window(tmp, buffer);
					return;
				}
			}
		}
		if (is_channel(who_from) && from_server != -1)
		{
			ChannelList *chan;
			if ((chan = (ChannelList *)find_in_list((List **)&(server_list[from_server].chan_list), 
				who_from, !USE_WILDCARDS)))
			{
				if (chan->window && from_server == chan->window->server)
				{
					add_to_window(chan->window, buffer);
					return;
				}
			}
		}
	}
	flag = 1;
	while ((tmp = traverse_all_windows(&flag)) != NULL)
	{
		if (((from_server == tmp->server) || (from_server == -1)) &&
		    (who_level & tmp->window_level))
		{
			add_to_window(tmp, buffer);
			return;
		}
	}
	if (from_server == curr_scr_win->server)
		tmp = curr_scr_win;
	else
	{
		flag = 1;
		while ((tmp = traverse_all_windows(&flag)) != NULL)
		{
			if (tmp->server == from_server)
				break;
		}
		if (!tmp)
			tmp = curr_scr_win;
	}
	add_to_window(tmp, buffer);
}

/*
 * update_all_windows: This goes through each visible window and draws the
 * necessary portions according the the update field of the window. 
 */
void update_all_windows()
{
	Window	*tmp;
	int	fast_window,
		full_window,
		r_status,
		u_status;

	for (tmp = current_screen->window_list; tmp; tmp = tmp->next)
	{
		if (tmp->display_size != tmp->old_size)
			resize_display(tmp);
		if (tmp->update)
		{
			fast_window = tmp->update & REDRAW_DISPLAY_FAST;
			full_window = tmp->update & REDRAW_DISPLAY_FULL;
			r_status = tmp->update & REDRAW_STATUS;
			u_status = tmp->update & UPDATE_STATUS;
			if (full_window)
				redraw_window(tmp, 1);
			else if (fast_window)
				redraw_window(tmp, 0);
			if (r_status)
				update_window_status(tmp, 1);
			else if (u_status)
				update_window_status(tmp, 0);
		}
		tmp->update = 0;
	}
	for (tmp = invisible_list; tmp; tmp = tmp->next)
	{
		if (tmp->display_size != tmp->old_size)
			resize_display(tmp);
		tmp->update = 0;
	}
	update_input(UPDATE_JUST_CURSOR);
	term_flush();
}

void close_all_screen(void)
{
	Screen *screen;

	for (screen = screen_list; screen && screen != current_screen;
			screen = screen->next)
	{
		if (screen->alive && screen->fdin != 0)
		{
			if (use_input)
				FD_CLR(screen->fdin, &readables);
			new_close(screen->fdin);
		}
	}
}


static	char	* next_line_back(Window *window)
{
#if NEW_NEXT_LINE_BACK
static	int	row = 0;
static	Lastlog	*LogLine;
	unsigned char	**TheirLines;
static	unsigned char	**ScreenLines = NULL;
	int size = 0, i, new;
	
 	if (!size)
 	{
 		new = MAXIMUM_SPLITS;
 		RESIZE(ScreenLines, char *, new);
 		while (size < new)
 			ScreenLines[size++] = 0;
 	}
	if (window)
	{
		LogLine = window->lastlog_head;
		row = -1;
	}
	if (row <= 0)
	{
		for (i = 0; ScreenLines[i]; i++)
			new_free(&ScreenLines[i]);

		if (!window && LogLine)
			LogLine = LogLine->next;
		if (!LogLine)
			return NULL;
		TheirLines = split_up_line(LogLine->msg);
		for (row = 0; TheirLines[row]; row++)
		{
 			if (row == size)
 			{
 				new = size + MAXIMUM_SPLITS;
 				RESIZE(ScreenLines, char *, new);
 				while (size < new)
 					ScreenLines[size++] = 0;
 			}
			ScreenLines[row] = TheirLines[row];
			TheirLines[row] = NULL;
		}
		if (window)
			return NULL;
	}
	return ScreenLines[--row];
#else
	static	int	row;
	static	Lastlog	*LogLine;
	unsigned char	**TheirLines;
static	unsigned char	*ScreenLines[MAXIMUM_SPLITS] = { NULL };

	if (window)
	{
		LogLine = window->lastlog_head;
		row = -1;
	}
	if (row <= 0)
	{
		for (row = 0; ScreenLines[row]; row++)
			new_free(&ScreenLines[row]);
		if (!window && LogLine)
			LogLine = LogLine->next;
		if (!LogLine)
			return NULL;
		TheirLines = split_up_line(LogLine->msg);
		for (row = 0; TheirLines[row]; row++)
		{
			ScreenLines[row] = TheirLines[row];
			TheirLines[row] = NULL;
		}
		if (window)
			return NULL;
	}
	return ScreenLines[--row];
#endif
}

static	void display_lastlog_lines(int start, int end, Window *window)
{
	Display	*Disp;
	char	*Line;
	int	i;

	next_line_back(window);

	for (i = window->new_scrolled_lines; i--;)
		next_line_back(NULL);

	for (i = 0, Disp = window->top_of_display; i < window->display_size; Disp = Disp->next, i++)
	{
		if (Disp->linetype)
			next_line_back(NULL);
	}
	
	for (i = 0; i < start; i++)
		next_line_back(NULL);

	for (; i < end; i++)
	{
		if (!(Line = next_line_back(NULL)))
			break;
		term_move_cursor(0, window->top + window->menu.lines + window->scrolled_lines - i - 1);
		rite(window, Line, 0, 0, 1, 0);
	}
}

static	void scrollback_backwards_lines(int ScrollDist)
{
	Window	*window;

	window = curr_scr_win;
	if (!window->scrolled_lines && !window->scroll)
	{
		term_beep();
		return;
	}

	window->scrolled_lines += ScrollDist;

	if (window->status_split)
		term_scroll(window->top + window->menu.lines, window->top + window->menu.lines + window->display_size - 1, -ScrollDist);
	else
		term_scroll(window->top, window->top - window->menu.lines + window->display_size, -ScrollDist);

	display_lastlog_lines(window->scrolled_lines - ScrollDist, window->scrolled_lines, window);
	cursor_not_in_display();
	update_input(UPDATE_JUST_CURSOR);
}

static	void scrollback_forwards_lines(int ScrollDist)
{
	Window	*window;

	window = curr_scr_win;
	if (!window->scrolled_lines)
	{
		term_beep();
		return;
	}
	if (ScrollDist > window->scrolled_lines)
		ScrollDist = window->scrolled_lines;

	window->scrolled_lines -= ScrollDist;

	if (window->status_split)
		term_scroll(window->top + window->menu.lines, window->top + window->menu.lines + window->display_size - 1, ScrollDist);
	else
		term_scroll(window->top, window->top - window->menu.lines + window->display_size, ScrollDist);

	if (window->scrolled_lines < window->display_size)
		redraw_window(window, ScrollDist + window->scrolled_lines - window->display_size);

	display_lastlog_lines(window->scrolled_lines-window->display_size,
		window->scrolled_lines - window->display_size + ScrollDist, window);

	cursor_not_in_display();
	update_input(UPDATE_JUST_CURSOR);

	if (!window->scrolled_lines)
	{
		window->new_scrolled_lines = 0;
		if (window->hold_mode)
			hold_mode(window, ON, 1);
		else
			hold_mode(window, OFF, 0);
	}
}

void scrollback_forwards(char key, char *ptr)
{
	scrollback_forwards_lines(curr_scr_win->display_size/2);
}

void scrollback_backwards(char key, char *ptr)
{
	scrollback_backwards_lines(curr_scr_win->display_size/2);
}


void scrollback_end(char key, char *ptr)
{
	Window	*window;

	window = curr_scr_win;
	window->new_scrolled_lines = 0;

	if (!window->scrolled_lines)
	{
		term_beep();
		return;
	}
	if (window->scrolled_lines < window->display_size)
		scrollback_forwards_lines(window->scrolled_lines);
	else
	{
		window->scrolled_lines = 0;
		redraw_window(window, 1);
		cursor_not_in_display();
		update_input(UPDATE_JUST_CURSOR);
		if (window->hold_mode)
			hold_mode(window, ON, 1);
		else
			hold_mode(window, OFF, 0);
	}
}

/*
 * scrollback_start: moves the current screen back to the start of
 * the scrollback buffer..  there are probably cases this doesn't
 * quite work.. -phone, april 1993.
 */
void scrollback_start(char key, char *ptr)
{
	Window	*window;

	window = curr_scr_win;
	if (!window->lastlog_size)
	{
		term_beep();
		return;
	}

	if (window->lastlog_size < window->display_size)
		scrollback_backwards_lines(window->lastlog_size);
	else
	{
		window->scrolled_lines = window->lastlog_size;
		display_lastlog_lines(window->scrolled_lines -
			window->display_size, window->scrolled_lines, window);
		cursor_not_in_display();
		update_input(UPDATE_JUST_CURSOR);
		window->new_scrolled_lines = 0;
		if (window->hold_mode)
			hold_mode(window, ON, 1);
		else
			hold_mode(window, OFF, 0);
	}
}

RETSIGTYPE sig_refresh_screen _((int unused))
{
	refresh_screen(0, NULL);
}

#if defined(WINDOW_CREATE) && !defined(WINNT)
extern	Window	*create_additional_screen _((void))
{
        Window  *win;
        Screen  *oldscreen;
        char    *displayvar,
                *termvar;
        int     screen_type = ST_NOTHING;
        struct  sockaddr_in NewSock;
        int     NsZ;
        int     s;
	fd_set	fd_read;
	struct	timeval	timeout;
	pid_t	child;
	int	old_timeout;
	unsigned short port;
	char buffer[BIG_BUFFER_SIZE + 1];


	if (!use_input)
		return NULL;


	/* Check for X first. */
	if ((displayvar = getenv("DISPLAY")))
	{
		if (!(termvar = getenv("TERM")))
			say("I don't know how to create new windows for this terminal");
		else
			screen_type = ST_XTERM;
	}
	/*
	 * Environment variable STY has to be set for screen to work..  so it is
	 * the best way to check screen..  regardless of what TERM is, the 
	 * execpl() for screen won't just open a new window if STY isn't set,
	 * it will open a new screen process, and run the wserv in its first
	 * window, not what we want...  -phone
	 */
	if (screen_type == ST_NOTHING && getenv("STY"))
		screen_type = ST_SCREEN;


	if (screen_type == ST_NOTHING)
	{
		say("I don't know how to create new windows for this terminal");
		return NULL;
	}

        say("Opening new %s...",
                screen_type == ST_XTERM ?  "window" :
                screen_type == ST_SCREEN ? "screen" :
                                           "wound" );
	port = 0;
	s = connect_by_number(NULL, &port, SERVICE_SERVER, PROTOCOL_TCP, 1);
	if (s < 0)
	{
		yell("Couldnt establish server side -- error [%d]", s);
		return NULL;
	}

	oldscreen = current_screen;
	set_current_screen(create_new_screen());

	switch ((child = fork()))
	{
		case -1:
		{
			kill_screen(current_screen);
			last_input_screen = oldscreen;
			set_current_screen(oldscreen);
			say("I couldnt fire up a new wserv process");
			break;
		}

		case 0:
		{
			char *opts, *of;
			char *xterm;
			char *args[64], **args_ptr = args;
			char geom[32];

			setuid(getuid());
			setgid(getgid());
		/*
		 * Unlike most other cases, it is important here to close down
		 * *ALL* unneeded file descriptors. Failure to do so can cause
		 * Things like server and DCC connections to fail to close on
		 * request. This isn't a problem with "screen", but is with X.
		 */
			close(s);
			close_all_screen();
			close_all_dcc();
			close_all_exec();
			close_all_server();

			if (screen_type == ST_SCREEN)
			{
				of = opts = m_strdup(get_string_var(SCREEN_OPTIONS_VAR));

				*args_ptr++ = "screen";
				while (opts && *opts)
					*args_ptr++ = new_next_arg(opts, &opts);
			}
			else if (screen_type == ST_XTERM)
			{
				snprintf(geom, 31, "%dx%d", CO+1, LI);
				of = opts = m_strdup(get_string_var(XTERM_OPTIONS_VAR));

				if (!(xterm = getenv("XTERM")))
					if (!(xterm = get_string_var(XTERM_VAR)))
						xterm = "rxvt";

				*args_ptr++ = xterm;
				*args_ptr++ = "-geometry";
				*args_ptr++ = geom;
				while (opts && *opts)
					*args_ptr++ = new_next_arg(opts, &opts);
				*args_ptr++ = "-e";
			}

			*args_ptr++ = WSERV_PATH;
			*args_ptr++ = "localhost";
			*args_ptr++ = ltoa((long)port);
			*args_ptr++ = NULL;
			new_free(&of);
			my_signal(SIGINT, SIG_IGN, 0);
			setsid();

			s = execvp(args[0], args);
			printf("aaaaaaaaaaa! [%d/%s]\n", s, strerror(errno));
			_exit(0);
		}
	}

	/* All the rest of this is the parent.... */
	NsZ = sizeof(NewSock);
	FD_ZERO(&fd_read);
	FD_SET(s, &fd_read);
	timeout.tv_sec = (time_t) 10;
	timeout.tv_usec = 0;
	sleep (1);

	/* using say(), yell() can be bad in this next section of code. */
	switch (select(NFDBITS , &fd_read, NULL, NULL, &timeout))
	{
	case -1:
	case 0:
	{
		int errnod = get_child_exit (child);
		close(s);
		kill_screen(current_screen);
		kill(child, SIGKILL);
		last_input_screen = oldscreen;
		set_current_screen(oldscreen);
                yell("child %s with %d", (errnod < 1) ? "signaled" : "exited",
                                         (errnod < 1) ? -errnod : errnod);
		yell("Errno is %d", errno);
		return NULL;
	}
	default:
	{
		current_screen->fdin = current_screen->fdout = accept(s, (struct sockaddr *) &NewSock, &NsZ);
		if (current_screen->fdin < 0)
			return NULL;
		FD_SET(current_screen->fdin, &readables);
		current_screen->fpin = current_screen->fpout = fdopen(current_screen->fdin, "r+");
		close(s);
		old_timeout = dgets_timeout(5);
		if (dgets(buffer, BIG_BUFFER_SIZE, current_screen->fdin, NULL) < 1)
		{
			FD_CLR(current_screen->fdin, &readables);
			close(current_screen->fdin);
			kill_screen(current_screen);
			kill(child, SIGKILL);
			last_input_screen = oldscreen;
			set_current_screen(oldscreen);
			(void) dgets_timeout(old_timeout);
			return NULL;
		}
		else
			malloc_strcpy(&current_screen->tty_name, buffer);

		win = new_window();
		(void) refresh_screen(0, NULL);
		set_current_screen(oldscreen);
		(void) dgets_timeout(old_timeout);
		return win;
	}
	}
	return NULL;
}
#endif /* WINDOW_CREATE */


void set_screens _((fd_set *rd, fd_set *wd))
{
Screen *screen;
	if (use_input)
		for (screen = screen_list; screen; screen = screen->next)
			if (screen->alive)
				FD_SET(screen->fdin, rd);
}



/* cough cough */
extern int key_pressed;

void do_screens _((fd_set *rd))
{
	Screen *screen;
	char buffer[BIG_BUFFER_SIZE + 1];

	if (!use_input)
		return;
	for (screen = screen_list; screen; screen = screen->next)
	{
		if (!screen->alive)
			continue;
		set_current_screen(screen);
		if (FD_ISSET(screen->fdin, rd))
		{
			/*
			 * This section of code handles all in put from the terminal(s).
			 * connected to ircII.  Perhaps the idle time *shouldn't* be 
			 * reset unless its not a screen-fd that was closed..
			 */
			time(&idle_time);
			if (cpu_saver)
			{
				cpu_saver = 0;
				status_update(1);
			}

			if (dumb)
			{
				int     old_timeout;

				old_timeout = dgets_timeout(1);
				if (dgets(buffer, INPUT_BUFFER_SIZE, screen->fdin, NULL))
				{
					(void) dgets_timeout(old_timeout);
					*(buffer + strlen(buffer) - 1) = '\0';
					if (get_int_var(INPUT_ALIASES_VAR))
						parse_line(NULL, buffer, empty_string, 1, 0);
					else
						parse_line(NULL, buffer, NULL, 1, 0);
				}
				else
				{
					bitchsay("BitchX exiting on EOF from stdin");
					irc_exit(1, "BitchX - EOF from stdin", NULL);
				}
				key_pressed = 13; /* *cough* *cough* */
			}
			else
			{
				int server;
				char	loc_buffer[BIG_BUFFER_SIZE + 1];
				int	n, i;

				server = from_server;
				from_server = get_window_server(0);
				last_input_screen = screen;
				if ((n = read(screen->fdin, loc_buffer, BIG_BUFFER_SIZE)) > 0)
				{
					extended_handled = 0;
					for (i = 0; i < n; i++)
						if (!extended_handled)
							edit_char(loc_buffer[i]);
					key_pressed = loc_buffer[0];
				}
#ifdef WINDOW_CREATE
				else if (screen != main_screen)
					kill_screen(screen);
#endif /* WINDOW_CREATE */
				else
					irc_exit(1, "Hey!  Where'd my controlling terminal go?", NULL);
				from_server = server;
			} 
		}
	} 
} 
